#
# serverFunctions.py
#
# by Arthur Kevin McGrath
#

"""
Functions that do things like create and delete 
servers and clusters.
General Comments
   NONE of the methods in this file that create things call AdminConfig.save()
   Therefore, no changes are committed unless YOU explicitly commit them
   Any methods that cause changes also call AdminConfig.queryChanges()
   so that you at least know which files are going to change if you
   commit the changes
"""
import AdminConfig,AdminControl,AdminTask,sys
import parsing


# common routines
# The methods that follow are NEVER expected to be called
# by code outside of this file

# whatFilesChanged
# Reports the configuration files that will be changed
# as a result of any code in this file
# The changes will take effect when you call AdminConfig.save()
def whatFilesChanged():
  print "The following congfiguration files have changes pending:"
  print AdminConfig.queryChanges()
  return


#setHeap()
#Changes the minimum heap size and
#maximum heap size of an array
#of Java Virtual Machines
#If the array is zero length, this method silently returns
#This method depends on a method in parsing.py
#parameters:
#   array - an array of Java Virtual Machine 
#           configuration IDs as returned by
#           AdminConfig.list()
#   min - the desired minimum heap size
#   max - the desired maximum heap size
#   diagnostic - set to 1 to turn verbose garbage collection on
#           set to zero (the default) to turn verbose garbage collection off
def setHeap( array, min, max, diagnostic=0 ):
  """setHeap()\nChanges the minimum heap size and maximum heap size\nof an array of Java Virtual Machines.  If the array is zero length, this method silently returns.  This method depends on a method in parsing.py\nparameters: (all parameters are mandatory)\n  array - an array of Java Virtual Machine\n           configuration IDs as returned by\n           AdminConfig.list()\n  min - the desired minimum heap size\n  max - the desired maximum heap size\n  diagnostic - set to 1 to turn verbose garbage collection on\n           set to zero (the default) to turn verbose garbage collection off"""
  verboseOrNot = 'false'
  if diagnostic != 0:
    verboseOrNot = 'true'
  #Diagnostic print
  #print array
  for x in array:
    #Diagnostic print
    #print x
    s = parsing.findValueInConfigurationIDString( x, 'server' )
    n = parsing.findValueInConfigurationIDString( x, 'node' )
    print AdminTask.setJVMProperties( '[-serverName ' + s + ' -nodeName ' +  n  + ' -verboseModeGarbageCollection ' + verboseOrNot +  ' -maximumHeapSize ' + str( max ) + ' -initialHeapSize ' + str( min ) + ' ]' )


#
# countConfigIDs()
#
def countConfigIDs( jaclList ):
  """
Counts the number of configuration IDs returned by an API such as AdminConfig.list()
parameters:
  jaclList -- a python String that contains a JACL list
returns:
  the number of configuration IDs on the list
  """
  pythonList = convertJACLListOfConfigIDsToPython( jaclList )
  # print len( pythonList )
  return len( pythonList )

#
# convertJACLListOfConfigIDsToPython( jaclList ):
#
def convertJACLListOfConfigIDsToPython( jaclList ):
  """
Converts a JACL list into a Python list of strings.
Many script methods -- such as AdminConfig.list() -- return JACL lists.
This method converts a JACL list of configuration IDs into a 
python list of strings
parameers:
  jaclList -- a python String that contains a JACL list
returns:
  a python list of strings.  Each string holds one configuration ID.
  """
  # change JACL list into Python string array

  # strip off leading and trailing square brackets
  intermediate = jaclList[1:-1 ]

  # change space delimiter to comma
  intermediate = intermediate.replace(  ") ",  ")," )

  # use the comma to delimit each string
  pythonList = intermediate.split( "," )
  return pythonList

# end of common routines

# Start of public routines

def createServer( nodeName, serverName ):
  """
create a server
parameters:
      nodeName - the node that hosts this server
      serverName - what do we want to call the new server?
  """
  print AdminTask.createApplicationServer( nodeName, "[-name " + serverName + " ]" )
  # whatFilesChanged()
  return

def deleteServer( nodeName, serverName ):
  """
delete a server
parameters:
      nodeName - the node that hosts this server
      serverName - which server in that node do we want?
  """
  AdminTask.deleteServer( "[-serverName " + serverName + " -nodeName " + nodeName + " ]" )
  whatFilesChanged()
  return

def changeServerPorts( nodeName = "  ", serverName = "  ", startingPortString = "  " ):
  """
Changes the addresses of the ports on this server.  After
this function returns, all ports will be in a consecutive block
of addresses.
parameters:
      nodeName - the node that hosts this server
      serverName - which server in that node do we want?
      startingPortString - an integer that will be the first
               of a block of consecutive ports for this server
returns:
      2 if the startingPortString is not numeric
  """
  nodeNameForList = "[-nodeName " + nodeName + "]"
  nodeNameForSet = "-nodeName " + nodeName + "  "
  if startingPortString.isdigit():
    pass
  else:
    print "Error:"
    print "   StartingPort MUST be numeric"
    return( 2 )

  startingPort = int( startingPortString )
  d = {}   # create an empty dictionary that we will populate in the loop that follows
  raw = AdminTask.listServerPorts( serverName, nodeNameForList )
  array = raw.splitlines()
  index = 0
  for x in array:
    #print "\n   -   -   -\n" + x
    rawList = x.split("[")
    middleList = []
    for s in rawList:
      #print "- - >" + s + "< - -"
      if len(s) > 2:
        middleList.append(s)
    # add an end point name, host and port value  to our dictionary
    d[ 'endPointName' ] = middleList[0]
    for x in range( 1, len(middleList) ):
      temp = middleList[x].split( "]" )
      #print temp
      nvArray = temp[0].split()
      #print nvArray
      d[ nvArray[0] ] = nvArray[1]
    #print d[ 'node' ] + "/" + d[ 'host' ]
    #print d[ 'endPointName' ], startingPort + index
    completePortParameters = " "
    if d[ 'host' ] == '*':
      completePortParameters = '[' + nodeNameForSet + ' -endPointName ' + d[ 'endPointName' ] + ' -host *  -port ' + str(startingPort + index) + ' -modifyShared true ]'
    else:
      completePortParameters = '[' + nodeNameForSet + ' -endPointName ' + d[ 'endPointName' ] + ' -host ' + d[ 'host' ] + ' -port ' + str(startingPort + index) + ' ]'
    #print serverName,completePortParameters
    print AdminTask.modifyServerPort( serverName,completePortParameters )
    index += 1

  # whatFilesChanged()
  return

def createCluster( nameOfCluster, createReplicationDomain = 0 ):
  """
Creates a cluster.  Does not add any member servers nor
does it deploy any applications to the cluster.
parameters:
   nameOfCluster
       String - the name of the cluster to create
   createReplicationDomain
       boolean - should we also create a replication domain for this cluster
       if no value is passed, the default is false
       In this version of Jython, 0 means false and 1 means true
returns:
   nothing
  """
  cluster = None
  # create the cluster
  print "  . . . creating a " + nameOfCluster + " cluster"
  if createReplicationDomain == 0:
    cluster = AdminTask.createCluster('[-clusterConfig [-clusterName ' + nameOfCluster + ']]')
  else:
    cluster = AdminTask.createCluster('[-clusterConfig [-clusterName ' + nameOfCluster + '] -replicationDomain [-createDomain true] ]')
  print cluster
  # whatFilesChanged()
  return


#
#
# createFirstClusterMember
#
def createFirstClusterMember( nameOfCluster, nodeName, serverName, joinReplicationDomain = 0, nodeGroupName = "DefaultNodeGroup", coreGroupName = "DefaultCoreGroup" ):
  """
Required for proper cluster member configuration.  This method
determines the core group and the node group that all members
of this cluster will be part of.
  """
  replicate = 'false'
  allClusters = AdminConfig.list( 'ServerCluster' ).splitlines()
  if len( allClusters ) < 1:
    print "There are no clusters named", nameOfCluster
    return

  thisCluster = ""
  for c in allClusters:
    if c.find( nameOfCluster ) > -1:
      break
  else:
    print "There are no clusters named", nameOfCluster
    return

  print "Adding first member " + nodeName + " / " + serverName + " to " + nameOfCluster
  if joinReplicationDomain != 0:
    replicate = 'true'
  print AdminTask.createClusterMember('[-clusterName ' + nameOfCluster +
        ' -memberConfig [-memberNode ' + nodeName + ' -memberName ' + serverName +
        ' -replicatorEntry ' + replicate   + '] ' +
        ' -firstMember[ -templateName default -nodeGroup ' + nodeGroupName +
        '  -coreGroup ' + coreGroupName + '   ] ' + ']')


#
# createAdditionalClusterMembers
#
def createAdditionalClusterMembers( nameOfCluster, listOfServerNames ):
  """
Adds cluster members to an existing cluster.  
REQUIRES:
  1. that all members be application servers AND
  2. that there already be at least one member of the cluster.
     (The first cluster member establishes:
        a. the type of server (application, proxy, etc)
        b. the node group each node in the cluster must belong to
        c. the core group each server in the cluster must belong to
None of the servers on the listOfServerNames may already exist.  This method will create them.
As long as the cluster has a replication domain, you may choose to include
any or all of your proposed new servers in that domain.
example
  The following adds three servers to an existing cluster.  One of the three servers
  will become part of an existing ReplicationDomain.  The other two will not
newClusterMembers = [ ( "node01","c01s01",1 ), ( "node01","c01s01",1 ), ( "node01","c01s01" ) ]
addClusterMembers( "clusterName", newClusterMembers )
parameters:
   nameOfCluster
       String - the name of the cluster to create
   listOfServerNames
       an array of tuples
       each tuple has:
           nodeName - (mandatory)
              String - the Node in which to create the server
           serverName - (mandatory)
              String - the for the newly created the server
           replicationDomain - (optional)
              boolean - should we create a replication domain
                        for the newly created server?
                        default value is false
  """
  allClusters = AdminConfig.list( 'ServerCluster' ).splitlines()
  if len( allClusters ) < 1:
    print "There are no clusters named", nameOfCluster
    return

  thisCluster = ""
  for c in allClusters:
    if c.find( nameOfCluster ) > -1:
      thisCluster = c
      break
  else:
    print "There are no clusters named", nameOfCluster
    return

  listOfProposedNewMemberServers = list( listOfServerNames )
  print "  . . . adding " + str( len(listOfProposedNewMemberServers) ) + " additional members to the " + nameOfCluster + " cluster"

  # add the rest of the cluster members
  for server in listOfProposedNewMemberServers:
    nodeName = server[0]
    serverName = server[1]
    replicate = "false"
    if len( server ) == 2:
      print AdminTask.createClusterMember('[-clusterName ' + nameOfCluster + ' -memberConfig [-memberNode ' + nodeName + ' -memberName ' + serverName  + ']]')
    if len( server ) == 3:
      # convert the ReplicationDomain parameter of our parameter list
      # into the 'true' or 'false' notation that AdminTask requires
      if server[2] == 0:
        replicate = "false"
      else:
        replicate = "true"
      print AdminTask.createClusterMember('[-clusterName ' + nameOfCluster + ' -memberConfig [-memberNode ' + nodeName + ' -memberName ' + serverName + ' -replicatorEntry ' + replicate + ']]')
  # whatFilesChanged()
  return


# deleteCluster
# This function will delete a cluster, all the members of a cluster
# and optionally, the replication domain associated with this cluster
# This function will work even if no servers are running.
# This function will run with conntype NONE
# parameters:
#    nameOfCluster
#        String - the name of the cluster to create
#    deleteReplicationDomain
#        boolean - should we also delete any replication domain
#        for this cluster
#        if no value is passed, the default is false
# throws:
#    com.ibm.websphere.management.cmdframework.CommandValidationException
#        if the cluster name cannot be found
def deleteCluster( nameOfCluster, deleteReplicationDomain = 0 ):
  """# deleteCluster\n# This function will delete a cluster, all the members of a cluster\n# and optionally, the replication domain associated with this cluster\n# This function will work even if no servers are running.\n# This function will run with conntype NONE\n# parameters:\n#    nameOfCluster\n#        String - the name of the cluster to create\n#    deleteReplicationDomain\n#        boolean - should we also delete any replication domain\n#        for this cluster\n#        if no value is passed, the default is false\n# throws:\n#    com.ibm.websphere.management.cmdframework.CommandValidationException\n#        if the cluster name cannot be found"""
  print AdminTask.help()
  command = ""
  if deleteReplicationDomain == 0:
    command = "[-clusterName \"" + nameOfCluster + "\"]"
    print command
  else:
    command = "[-clusterName \"" + nameOfCluster + "\" -replicationDomain [-deleteRepDomain true] ]"
    print command
  AdminTask.deleteCluster( command )
  whatFilesChanged()
  return

# setVariable
#  will create a new variable if none exists at the requested scope.
# If a variable already exists at the requested   scope, its value will be changed
# parameters
#   varName
#       String
#   varScope
#       String - this must be a value that AdminConfig.getid() can accept
#   newValue
#       String
#   newDescription
#       String - optional
def setVariable( varName, varScope, newValue, newDescription = "" ):
  scopeID = AdminConfig.getid( varScope )
  vlistRaw = AdminConfig.list( "VariableSubstitutionEntry", scopeID )
  #print "- - >" + vlistRaw + "<- - length is", len(vlistRaw)
  if len(vlistRaw) > 0:
    vlist = vlistRaw.splitlines()
  else:
    setNewVariable( varName, varScope, newValue, newDescription )
    return

  for v in vlist:
    n = AdminConfig.showAttribute( v, "symbolicName" )
    if n == varName:
      print "     -   -   -"
      print AdminConfig.show(v)
      AdminConfig.modify( v, [ ["value", newValue] ] )
      break
  else:
    setNewVariable( varName, varScope, newValue, newDescription )
    return

  whatFilesChanged()
  return

def setNewVariable( varName, varScope, newValue, newDescription ):
  #AdminTask uses a different format for scope, so we have to massage
  #our scope into that format
  s = varScope.replace( "/"," " )
  s = s.strip()                      # remove leadding and trailing spaces
  s = s.replace(":","=")      # the   colon becomes =
  reformattedScope = s.replace( " ", "," )   # commas separate levels of scope
  AdminTask.setVariable('[-scope ' + reformattedScope + '      -variableName ' + varName + '    -variableValue ' + newValue + '      -variableDescription \"' + newDescription + '\"]')
  whatFilesChanged()
  return


# getVariablesInScope
# return a Dictionary with all the variables in a given scope
# parameters:
#       scope
#               String in the form of Node=xxx,Server=yyy
#       varName
#               String optional parameter
#                       if present, the name of the WAS variable
#                       if not present, return all WAS variables in the specified scope
def getVariablesInScope( scope, varName = "@!~~$%^^" ):
  raw = ""
  result = {}
  if varName == "@!~~$%^^":
    try:
      raw = AdminTask.showVariables( "[ -scope " +  scope +  " ]"  )
    except:
      raw = ""
  else:
    try:
      raw = AdminTask.showVariables( "[ -scope " +  scope +  "  -variableName  " + varName +" ]"  )
    except:
      raw = ""
  # convert the string into a proper python List
  raw = raw.replace( "] [", "],[" )
  #strip off outtermost square brackets
  if  len(raw) > 4:
    raw = raw[2:-2]
  else:
    return result
  # convert each WAS variable into a List
  outterList = raw.split(",")
  for o in outterList:
    #make each variable into a key:value
    #strip off the square brackets that surround each key:value pair
    str = o[1:-1]
    #separate the key from the value
    lst = str.split( " ", 1)
    # and add it to our master List
    result[ lst[0] ]  =  lst[1]
  
  return result


# displayListofVariables
# calls getVariablesInScope() and displays the result in a human readable form
# parameters:
#       scope
#               String in the form of Node=xxx,Server=yyy
def displayListOfVariables( scope ):
  vDict = getVariablesInScope(scope)
  if len(vDict) > 0:
    keyLst = vDict.keys()
    keyLst.sort()
    for k in keyLst:
        print k," = ",vDict[k]
  else:
    print "nothing to display"

# createTransportChain
#
# parameters:
#       serverName - name of the server that will listen on this port
#       chainName - this is both the name of the chain we create and a stub for the name of the end point that connects to the chain
#       hostName - gee, what could this be?
#       portNumber - ditto
# returns:
#       the configID of the newly created transport chain
def createTransportChain(  serverName,chainName, hostName, portNumber ):
  tcpChannelService = AdminConfig.list( 'TransportChannelService',  AdminConfig.getid( '/Server:' + serverName + '/' )  )
  print tcpChannelService
  print "  -  -  -  -"
  print AdminTask.createTCPEndPoint( '\'' + tcpChannelService + '\'',  '\'[-name ' + chainName + '_END_POINT' + '  -host ' + hostName + '  -port ' + portNumber + ' ]\' ' )
  # the template below is something of a magic incantation
  # I got it from looking at Command Assistance from the Admin Console
  # A better way is to type print AdminConfig.listTemplates( 'Chain' )
  # There is a plain chain and a secure chain
  chain = AdminTask.createChain( tcpChannelService, '\'[ -template WebContainer(templates/chains|webcontainer-chains.xml#Chain_1 -name ' + chainName + ' \' ' )
  print chain
  whatFilesChanged()
  return


# displayVirtualHosts
# Either:
#     shows all the VirtualHosts in a cell
# Or
#     shows all the VirtualHosts in the Node of your choice
# Parameters:
#   parentNodeName - (optional) the name of a Node
#         This meeethod will display all 
def displayVirtualHosts( parentNodeName = "@#@#@#"  ):
  raw = ""
  if cmp(parentNodeName, "@#@#@#") == 0:
    raw = AdminConfig.list( 'VirtualHost' )
  else:
    raw = AdminConfig.list( 'VirtualHost', AdminConfig.getid( "/Node:" + parentNodeName   + "/" )  )
  arbitraryLowerLimit = 5
  if len(raw) < arbitraryLowerLimit:
    print "\nThere are no VirtualHosts configured anywhere in this cell"
    return
  virtualHostArray = raw.splitlines()
  for a in virtualHostArray:
    print "  *  *  *  *  *  *  * name of VirtualHost: ", AdminConfig.showAttribute( a, "name" ), "  *  *  *  *  *  *  *"
    print a
    # replace some junk that IBM put in the output with spaces
    # remove the spaces
    # convert the result into an array
    # walk the array
    aliases = AdminConfig.showAttribute( a, "aliases" )
    aliases = aliases.replace( '[', ' ' )
    aliases = aliases.replace( ']', ' ' )
    aliases = aliases.strip()
    aliasArray = aliases.split()
    for b in aliasArray:
      print AdminConfig.showAttribute( b, 'hostname' ), ":", AdminConfig.showAttribute( b, 'port' )

# createVirtualHost
# creates a new VirtualHost at either the Cell level or at the Node level
# parameters:
#       parentID - the configID of the parent.  MUST be either a Cell or a Node
#       vhName - the name for this new VirtualHost
#       aliasList - a python List of Lists of host aliases.
#           each host alias must have a host name and a port value
#                 [  ['hostname', 'newsflash.cnn.com'], ['port', '66993']  ]
#           You can have as many aliases as you want
#                 [     [  ['hostname', 'newsflash.cnn.com'], ['port', '66993']  ],
#                       [  ['hostname', 'securenewsflash.cnn.com'], ['port', '66983']  ],
#                       [  ['hostname', 'www.cnn.com'], ['port', '38080']  ]      ]
# returns:
#     the configID of the newly created VirtualHost
def createVirtualHost( parentID, vhName, aliasList ):
  result = "something   went wrong"
  #start of parameterList
  parameterList = [  ['name', vhName ] ,[ 'aliases', aliasList ] ]
  try:
    result = AdminConfig.create( 'VirtualHost', parentID, parameterList )
    print result
  except:
    print "\n  .  .  .  .  Something went wrong.  We passed the following ..."
    print parentID
    print parameterList
    print "\n.  .  .  .  .  and this went wrong"
    print sys.exec_info
    print "  .  .  .  .  .  .end of error  .  .  .  ."
  whatFilesChanged()
  return


#setServerHeaps()
#Set min and max heap size for all servers of a given type
#If no server type is specified, set min and max heap size
# for all servers in a cell.  In addition, you can enable or disable
# verbose garbage collection for either a class of servers or
# for all servers in a cell
#parameters:
#    minHeap - the minimum heap size you want (in MB).  This size MUST be smaller than max
#    maxHeap - the maximum heap size you want (in MB).  This size MUST be larger than min
#    serverClass - OPTIONAL PARAMETER - a String that names the kind of server you want to modify.  Must be a type returned by AdminConfig.types().  All servers of that type will be modified.  If no servers of that type exist in your cell, this routine silently returns without doing anything
#    verbose - OPTIONAL PARAMETER - set to 1 if you want to enable verbose garbage collection.  Set to zero if you want to disable verbose garbage collection.  The default is zero.
def setServerHeaps( minHeap, maxHeap, serverClass, verbose):
  "Set min and max heap size for all servers of a given type.  If no server type is specified, set min and max heap size for all servers in a cell.  In addition, you can enable or disable verbose garbage collection for either a class of servers or for all servers in a cell\nparameters:\n    minHeap - the minimum heap size you want (in MB).  This size MUST be smaller than max\n    maxHeap - the maximum heap size you want (in MB).  This size MUST be larger than min\n    serverClass - OPTIONAL PARAMETER - a String that names the kind of server you want to modify.  Must be a type returned by AdminConfig.types().  All servers of that type will be modified.  If no servers of that type exist in your cell, this routine silently returns without doing anything\n    verbose - OPTIONAL PARAMETER - set to 1 if you want to enable verbose garbage collection.  Set to zero if you want to disable verbose garbage collection.  The default is zero."
  raw = AdminConfig.list( serverClass )
  if len(raw) > 5:
      n = raw.splitlines()
      setHeap( n, minHeap, maxHeap, verbose )
  whatFilesChanged()
  return


#synchronizationConfiguration()
#configure the terms of the Node Synchronization service
#parameters:
#    id - String - configuration ID of a ConfigSynchronizationService for one node
#    interval - integer - how long in minutes between node synchronizations
#    synchOnStartup - OPTIONAL PARAMETER - set to zero to prevent node synchronization on server startup.  Set to 1 to enable.   Default value is zero
def synchronizationConfiguration( id, interval, synchOnStartup=0 ):
  "synchronizationConfiguration()\nconfigure the terms of the Node Synchronization service\nparameters:\n    id - String - configuration ID of a ConfigSynchronizationService for one node\n    interval - integer - how long in minutes between node synchronizations\n    synchOnStartup - OPTIONAL PARAMETER - set to zero to prevent node synchronization on server startup.  Set to 1 to enable.   Default value is zero"
  synch = 'false'
  if synchOnStartup == 1:
    synch = 'true'
  AdminConfig.modify( id, [ ['synchOnServerStartup', synch] ] )
  AdminConfig.modify( id, [ ['synchInterval', interval] ] )
  whatFilesChanged()
  return


#checkSync()
#returns the synchronization status for each node in the cell.
#This method can only check nodes if their node agent is up
def checkSync():
  """returns the synchronization status for each node in the cell.  
This method can only check nodes if their node agent is up
  """
  ns = []
  n = "garbage"
  serverName = "garbage"
  raw = AdminControl.queryNames( 'type=NodeSync,*' )
  if len(raw) > 5:  #arbitrary minimum string length
    ns = raw.splitlines()
    for x in ns:
      # DIAGNOSTIC print
      #print x
      # show node name
      n = parsing.findValueInObjectNameString( x, 'node' )
      serverName = parsing.findValueInObjectNameString( x, 'process' )
      # DIAGNOSTIC print
      #print n
      syncStatus = 'has NOT been synchronized'
      # show current synchronization status
      if AdminControl.invoke( x, 'isNodeSynchronized' ) == 'true':
        syncStatus = 'is synchronized'
      print n,syncStatus, " The last attempt to synchronize",n,"was:"
      # show last attempted synchronization
      print AdminControl.invoke( x, 'getResult' )
      print "The configuration settings for the ConfigSynchronizationService for",n,"are:"
      enclosingScope = AdminConfig.getid( '/Node:' + n + '/Server:' + serverName + '/' )
      cs = AdminConfig.list( 'ConfigSynchronizationService', enclosingScope)
      print AdminConfig.show(cs)
  return

#syncAllNodes()
#synchronizes any node whose node agent is up
def syncAllNodes():
  "Synchronizes any node whose node agent is up"
  raw = AdminControl.queryNames( 'type=NodeSync,*' )
  if len(raw) > 5:  #arbitrary minimum string length
    ns = raw.splitlines()
    for x in ns:
      AdminControl.invoke( x, 'sync' )
      print AdminControl.invoke( x, 'getResult' )



# main method starts here
if __name__ == "main":
  print "Running standalone . . ."
else:
  "A bunch of methods that do useful things to servers"
  print  __name__ , "module is loaded"
  #print __name__,"can see "
  #print dir()
